# System.Text.Encoding

The **Encoding** is used to decode text to unicode code points or encode the unicode code points to text.

The **Encoding** is a prototype, so you can't used it to create objects, but can use it to generate encoder prototypes.

The **Encoding** is a validatable type, but the encoder prototypes generated from it aren't validatable types, but they can pass the validation of the **Encoding**, so you can still use it like:

```lua
__Arguments__{ String, System.Text.Encoding/System.Text.UTF8Encoding }
function getCodes(str, encoder) end
```


## System.Text.ASCIIEncoding

The **ASCIIEncoding** is used to provide the support for ASCII encoding:

```lua
require "PLoop" (function(_ENV)
	import "System.Text"

	-- 116, 101, 115, 116
	print(List( ASCIIEncoding.Decodes("test") ):Join(", "))

	-- test
	print( ASCIIEncoding.Encodes{ 116, 101, 115, 116 } )

	-- test
	print( ASCIIEncoding.Encodes(ASCIIEncoding.Decodes("test")) )

	-- 116
	print(ASCIIEncoding.Decode("test"))

	-- 101
	print(ASCIIEncoding.Decode("test", 2))

	-- t
	print(ASCIIEncoding.Encode(116))

	-- e
	print(ASCIIEncoding.Encode(101))
end)

```

The encoder prototypes all have provided four methods:

Method         |Arguments                    |Description
:--------------|:----------------------------|:-------------------
Encode         |code:Number                  |Encode the unicode code point to a char
Encodes        |list:Table + IList           |Encode all codes in the list to a text
Encodes        |iter:Function, ...:Any * 0   |Encode all codes from the iterator to a text
Decode         |str:String, index:Number/1   |Decode a char at the index of the text to unicode code point, also will return the byte count, nil means 1-byte
Decodes        |str:String                   |Return an iterator to generate all unicode code points from the target text


## System.Text.UTF8Encoding

The **UTF8Encoding** is used to provide the support for UTF8 encoding


## System.Text.UTF16EncodingLE

The **UTF16EncodingLE** is used to provide the support for UTF-16 encoding with little-endian.


## System.Text.UTF16EncodingBE

The **UTF16EncodingBE** is used to provide the support for UTF-16 encoding with big-endian.


## Define an encoder prototype

To define the encoder prototype, only need provide the **Encode** and **Decode** functions, take the ASCIIEncoding as an example:

```lua
System.Text.Encoding "ASCIIEncoding" {
	encode                  = string.char,
	decode                  = string.byte,
}
```

If only provide the name of the prototype, it'll be saved to **System.Text** namespace, the **ASCIIEncoding** is using *string.char* as **Encode** and the *string.byte* as **Decode**, so you can create your own encoder like it.

But keep in mind, the **Decode** should also return the byte count as the second return value, if omit, the system will use 1 as default byte count.
